"""

OBSOLETE:

    - keeping so we can incorporate infomap aggregation strategy



"""


import numpy as np
import networkx as nx
from collections import Counter

from .labelers import AttributeLabeler


class InfomapClassifier:
    """
    Classifier for assigning labels to communities.

    Attributes:

        classifier (CellClassifier) - individual cell classifier

    """

    def __init__(self, cells, cell_classifier, rule='weighted'):
        """
        Instantiate community classifier.

        Args:

            cells (pd.DataFrame) - cell data including community labels

            cell_classifier (BayesianClassifier) - labels individual cells

            rule (str) - voting rule, e.g. 'weighted' or 'majority'

        """
        self.classifier = self.build_classifier(cells, cell_classifier, rule)

    def __call__(self, communities):
        """ Classify communities. """
        return self.classifier(communities)

    @classmethod
    def from_layer(cls, layer, cell_classifier, **kwargs):
        """ Instantiate community classifier from a layer. """
        return cls.__class__(layer.data, cell_classifier, **kwargs)

    @staticmethod
    def get_mode(x):
        """ Returns most common value in an array. """
        mode, count = Counter(x).most_common(1)[0]
        return mode

    @classmethod
    def build_voter(cls, cell_classifier, rule='proportional'):
        """
        Build voting function.

        Args:

            cell_classifier (BayesianClassifier) - labels individual cells

            rule (str) - voting rule, e.g. 'weighted' or 'majority'

        """

        # aggregate votes via majority rules
        if rule == 'majority':
            def voter(x):
                return cls.get_mode(cell_classifier(x))

        # aggregate maximum mean posterior (proportional representation)
        elif rule == 'proportional':
            def voter(x):
                posterior = cell_classifier.evaluate_posterior(x)
                return posterior.mean(axis=0).argmax()

        # aggregate votes weighted by posterior probability of each label
        elif rule == 'weighted':
            def voter(x):
                posterior = cell_classifier.evaluate_posterior(x)
                confidence = posterior.max(axis=1)
                genotypes = posterior.argmax(axis=1)
                ind = np.argsort(genotypes)
                starts = np.searchsorted(genotypes[ind], np.arange(4))
                lengths = np.diff(starts)
                return np.argmax([confidence[ind][slice(s, s+l)].sum() for s, l in zip(starts[:-1], lengths)])

        else:
            raise ValueError('Voter rule not recognized.')

        return voter

    @classmethod
    def build_classifier(cls, cells, cell_classifier, rule='weighted'):
        """
        Build classifier assigning genotypes to graph communities.

        Args:

            cells (pd.DataFrame) - cell data including community labels

            cell_classifier (BayesianClassifier) - labels individual cells

            rule (str) - voting rule, e.g. 'weighted' or 'majority'

        """
        #majority_vote = lambda x: cls.get_mode(cell_classifier(x))
        voter = cls.build_voter(cell_classifier, rule=rule)
        communities = cells.groupby('community')
        community_to_genotype = communities.apply(voter).to_dict()
        community_to_genotype[-1] = -1
        return np.vectorize(community_to_genotype.get)


class InfomapLabeler(AttributeLabeler):
    """
    Object for assigning genotypes to cells based on their local community.

    Attributes:

        graph (Graph) - graph connecting adjacent cells

        cell_classifier (CellClassifier) - callable object

        labeler (InfomapClassifier) - callable object

    Inherited attributes:

        label (str) - name of label field to be added

        attribute (str) - existing cell attribute used to determine labels

    """

    def __init__(self, graph, cell_classifier,
                 label='community_genotype',
                 attribute='community',
                 twolevel=False,
                 **kwargs):
        """
        Instantiate community-based genotype annotation object.

        Args:

            graph (Graph) - graph connecting adjacent cells

            cell_classifier (CellClassifier) - callable object

            label (str) - name of <genotype> attribute to be added

            attribute (str) - name of attribute defining community affiliation

            twolevel (bool) - if True, perform two-level clustering

            kwargs: keyword arguments for InfomapClassifier

        """

        # store label and attribute field names
        self.label = label
        self.attribute = attribute

        # run community detection and store graph
        graph.find_communities(twolevel=twolevel)
        self.graph = graph

        # store cell classifier
        self.cell_classifier = cell_classifier

        # build genotype labeler based on community classifier
        self.labeler = self.build_classifier(**kwargs)

    @classmethod
    def from_layer(cls, layer, **kwargs):
        """
        Instantiate from layer.

        Args:

            layer (Layer)

        Returns:

            labeler (InfomapLabeler derivative)

        """
        return cls.__class__(layer.graph, layer.classifier, **kwargs)

    def build_classifier(self, **kwargs):
        """
        Build community classifier.

        Args:

            kwargs: keyword arguments for community classifier

        Returns:

            classifier (func) - maps communities to labels

        """

        # assign community labels
        self.graph.data['community'] = -1
        ind = self.graph.nodes
        self.graph.data.loc[ind, 'community'] = self.graph.community_labels

        # build community classifier
        classifier = InfomapClassifier(self.graph.data,
                                         self.cell_classifier,
                                         **kwargs)

        return classifier


class KatzLabeler(InfomapLabeler):
    """
    Object for assigning genotypes to cells based on their local community by using the Katz centrality of the posterior distributions fit to cell fluorescence levels.

    Attributes:

        graph (Graph) - graph connecting adjacent cells

        cell_classifier (CellClassifier) - callable object

        labeler (KatzClassifier) - callable object

    Inherited attributes:

        label (str) - name of label field to be added

        attribute (str) - existing cell attribute used to determine labels

    """

    def __init__(self, graph, cell_classifier,
                 label='katz_genotype',
                 **kwargs):
        """
        Instantiate community-based genotype annotation object.

        Args:

            graph (Graph) - graph connecting adjacent cells

            cell_classifier (CellClassifier) - callable object

            label (str) - name of <genotype> attribute to be added

            kwargs: keyword arguments for KatzClassifier, e.g. alpha

        """

        # store label and attribute field names
        self.label = label

        # run community detection and store graph
        self.graph = graph

        # store cell classifier
        self.cell_classifier = cell_classifier

        # build genotype labeler based on community classifier
        self.labeler = self.build_classifier(**kwargs)

    def assign_labels(self, data):
        """
        Assign labels by adding <label> field to cell measurement data.

        Args:

            data (pd.DataFrame) - cells measurement data

        """
        data[self.label] = self.labeler(data.index.values)

    def build_classifier(self, **kwargs):
        """
        Build Katz classifier.

        Args:

            kwargs: keyword arguments for Katz classifier, e.g. alpha

        Returns:

            classifier (func) - maps measurement index to cell genotype

        """
        return self._build_classifier(self.graph, self.cell_classifier, **kwargs)

    @staticmethod
    def _build_classifier(graph, cell_classifier, alpha=0.9):
        """
        Construct classifier baed on the maximum Katz centrality of posterior distributions applied to undirected edges weighted by node similarity.

        Args:

            graph (Graph) - graph connecting adjacent cells

            cell_classifier (BayesianClassifier) - labels individual cells

            alpha (float) - attenuation factor

        Returns:

            genotypes_dict (dict) - maps measurement index to genotype

        """

        # build undirected graph weighted by node similarity
        G = graph.get_networkx()

        # evaluate posterior genotype distribution for each node
        posterior = cell_classifier.evaluate_posterior(graph.df.loc[list(G.nodes)])

        # compile normalized adjacency matrix
        adjacency = nx.to_numpy_array(G)
        adjacency /= adjacency.sum(axis=0)

        # evaluate centrality
        n = np.array(adjacency).shape[0]
        centrality = np.linalg.solve(np.eye(n, n)-(alpha*adjacency), (1-alpha)*posterior)

        # build classifier that maps model distributions to genotypes.
        #get_label = np.vectorize(cell_classifier.component_to_label.get)
        node_labels = centrality.argmax(axis=1)

        # return genotype mapping
        index_to_genotype = dict(zip(list(G.nodes), node_labels))

        return np.vectorize(index_to_genotype.get)



from copy import deepcopy
import numpy as np
import matplotlib.pyplot as plt
from scipy.ndimage import binary_dilation, generate_binary_structure
from matplotlib.gridspec import GridSpec

from ..visualization.masking import Mask


class BackgroundExtraction:
    """
    Object for extracting image background pixels.

    Attributes:

        layer (Layer) - layer RGB image

        bg_mask (np.ndarray[bool]) - background mask, True where background

    """

    def __init__(self, layer, niters=10):
        """
        Instantiate background extraction.

        Args:

            layer (Layer) - layer RGB image

            niters (int) - number of binary foreground dilations

        """

        # store layer
        self.layer = layer

        # build background mask
        self.bg_mask = self.build_background_mask(niters)

    def build_background_mask(self, niters=0):
        """
        Construct background mask by dilating foregound.

        Args:

            niters (int) - number of binary dilations

        Returns:

            bg_mask (np.ndarray[bool]) - background mask, True where background

        """

        # re-run image preprocessing to obtain foreground threshold
        seg_params = self.layer.metadata['params']['segmentation_kw']

        if 'preprocessing_kws' in seg_params.keys():
            preprocessing_kws = seg_params['preprocessing_kws']
        else:
            preprocessing_kws = None

        if preprocessing_kws is not None:
            bg = self.layer.get_channel(self.layer.metadata['bg'])
            _ = bg.preprocess(**preprocessing_kws)
            bg.set_otsu_mask()
            bg_mask = bg.mask
        else:
            bg_mask = (self.layer.labels!=0)

        # dilate foreground
        bg_mask = self.dilate_foreground(bg_mask, niters)

        return bg_mask

    @staticmethod
    def dilate_foreground(foreground, niters=5):
        """ Apply binary dilation to foreground mask. """
        struct = generate_binary_structure(2, 2)
        fg_mask = binary_dilation(foreground, struct, niters)
        bg_mask = ~fg_mask
        return bg_mask

    def isolate_pixels(self, channel):
        """
        Isolate pixels in image background.

        Args:

            channel (int) - color channel to be extracted

        Returns:

           px (np.ma.masked_array) - background masked pixel intensities

        """
        px = self.layer.get_channel(channel).im
        return np.ma.masked_array(px, ~self.bg_mask)

    def extract_pixels(self, channel):
        """
        Extract pixels from image background.

        Args:

            channel (int) - color channel to be extracted

        Returns:

            px (np.ndarray[float]) - 1D array of background pixel intensities

        """
        bg_px = self.isolate_pixels(channel)
        return bg_px[~bg_px.mask].data

    def plot_foreground_mask(self, invert=False, ax=None, figsize=(3, 3)):
        """
        Plot foreground mask.

        Args:

            invert (bool) - if True, mask background rather than foreground

            ax (matplotlib.axes.AxesSubplot) - if None, create figure

            figsize (tuple) - figure size

        Returns:

            figure

        """

        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        else:
            fig = plt.gcf()

        # TO DO: handle alternate channel specification... non RGB... etc
        # extract and visualize red/green channels
        rg = deepcopy(self.layer.im)
        rg[:,:,-1] = 0
        ax.imshow(rg)

        # add foreground mask
        if invert:
            mask = Mask(~self.bg_mask)
        else:
            mask = Mask(self.bg_mask)

        mask.add_contourf(ax, alpha=0.5, hatches=['//'])
        mask.add_contour(ax, lw=2, color='w')
        ax.axis('off')

        return fig



from ..annotation import WeightedGraph
from ..annotation import Annotation


class Training:
    """ Methods for graph construction and training an Annotation object. """

    @staticmethod
    def build_graph(measurements,
                    weighted_by='clonal_marker',
                    logratio=True):
        """
        Returns WeightedGraph object.

        Args:

            measurements (pd.DataFrame) - measurement data

            weighted_by (str) - attribute used to weight graph edges

            logratio (bool) - if True, weight by log-ratio of attribute level

        Returns:

            graph (spatial.WeightedGraph)

        """
        return WeightedGraph(measurements,
                             xykey=['x', 'y'],
                             weighted_by=weighted_by,
                             logratio=logratio)

    @staticmethod
    def train(*graphs, attribute='clonal_marker', **kwargs):
        """
        Train an Annotation model on the measurements in this layer then return the optimal model.

        Args:

            graphs (collection of spatial.WeightedGraph) - training data

            attribute (str) - measured attribute used to determine labels

            kwargs: keyword arguments for Annotation, including:

                sampler_type (str) - either 'radial', 'neighbors', 'community'

                sampler_kwargs (dict) - keyword arguments for sampler

                min_num_components (int) - minimum number of mixture components

                max_num_components (int) - maximum number of mixture components

                addtl_kwargs: keyword arguments for Classifier

        Returns:

            annotator (Annotator object) - annotator fit to training data

        """
        annotator = Annotation(attribute, **kwargs)
        selector = annotator.train(*graphs)
        return annotator



import sys, os
from subprocess import call
from distutils.core import setup
from distutils.command.install import install as _install
from setuptools import find_packages
from os import path

# read the contents of your README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='flyqma',
    version='0.4',
    author='Sebastian Bernasek',
    author_email='sebastian@u.northwestern.edu',
    packages=find_packages(exclude=('tests', 'scripts', 'validation')),
    scripts=[],
    url='https://sbernasek.github.io/flyqma/',
    license='MIT',
    description='Quantitative mosaic analysis of Drosophila imaginal discs.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    python_requires='>=3',
    install_requires=[
        "tifffile >= 0.15.0",
        "scipy >= 1.4.0",
        "scikit-image >= 0.14.0",
        "scikit-learn >= 0.19.2",
        "statsmodels >= 0.11.0",
        "pandas >= 0.23.3",
        "tables >= 3.4.4",
        "seaborn >= 0.9.0",
        "networkx >= 2.1",
        "dill >= 0.2.8.2",
        "joblib >= 0.9.0b4",
        "pyyaml",
        #"cython >= 0.22.1",
        #"pillow >= 5.2.0",
        #"infomap >= 1.0.0b8"
    ],
    tests_require=['nose'],
    test_suite='nose.collector'
)



import numpy as np
from scipy.signal import argrelextrema
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap, Normalize

from ...visualization import *


class MixtureVisualization:
    """ Methods for visualizing a mixture-model based classifier. """

    @property
    def label_colors(self):
        """ RGB color for each class label. """
        component_to_label = np.vectorize(self.component_to_label.get)
        labels = component_to_label(np.arange(self.num_components))
        label_colors = self.cmap(labels)[:, :-1]
        return label_colors

    @property
    def support(self):
        """ Model support. """
        return self.model.support

    @property
    def esupport(self):
        """ Empirical support vector (sorted values). """
        return np.sort(self.values, axis=0)

    @property
    def epdf(self):
        """ Empirical PDF over support. """
        num_bins = self.num_samples // 25
        bins = np.linspace(self.support.min(), self.support.max(), num_bins)
        counts, edges = np.histogram(self.values, bins=bins, normed=True)
        bin_centers = [(edges[i]+edges[i+1])/2. for i in range(len(edges)-1)]
        return edges[:-1], counts

    @property
    def ecdf(self):
        """ Empirical CDF over support. """
        return np.linspace(0, 1, len(self.values), endpoint=False)

    @property
    def pdf(self):
        """ Model PDF over support. """
        return self.model.pdf

    @property
    def component_pdfs(self):
        """ Weighted component PDFs over support. """
        return self.model.component_pdfs

    @property
    def component_cdfs(self):
        """ Returns weighted CDF of each component over support. """
        cdfs = np.vstack([x.cdf(self.support) for x in self.model.components])
        cdfs *= self.model.weights_.reshape(-1, 1)
        return cdfs

    @property
    def support_labels(self):
        """ Labels for support vector. """
        return self.classifier(self.support.reshape(-1, 1))

    @default_figure
    def plot_pdf(self,
                  density=1000,
                  alpha=0.5,
                  xmin=None,
                  xmax=None,
                  ymin=None,
                  ymax=None,
                  ax=None):
        """
        Plot model density function, colored by output label.
        """

        # plot model pdf segments, colored by output label
        support_labels = self.support_labels
        breakpoints = [0]+list(np.diff(support_labels).nonzero()[0]+1)+[None]
        for i, bp in enumerate(breakpoints[:-1]):
            indices = slice(bp, breakpoints[i+1])
            segment_support = self.support[indices]
            segment_pdf = self.pdf[indices]
            segment_labels = support_labels[indices]
            segment_color = self.cmap(segment_labels)
            ax.fill_between(segment_support, segment_pdf, color=segment_color)

        # plot overall model pdf
        ax.plot(self.support, self.pdf, '-', c='k', lw=2)

        # format axis
        if ymax is None:
            maxima = self.pdf[argrelextrema(self.pdf, np.greater)]
            ymax = 2.5*np.product(maxima)**(1/maxima.size)

        ax.set_xlim(self.model.lbound, self.model.ubound)
        ax.set_ylim(0, ymax)
        ax.set_xlabel('Values', fontsize=8)
        ax.set_ylabel('Density', fontsize=8)

    @default_figure
    def plot_pdfs(self,
                  empirical=False,
                  line=True,
                  fill=True,
                  density=1000,
                  alpha=0.5,
                  cmap=None,
                  vmin=-1,
                  xmax=None,
                  ymax=None,
                  ax=None):
        """
        Plot density function for each distribution, colored by output label.

        Args:

            ax (matplotlib.axes.AxesSubplot) - if None, create figure

            empirical (bool) - if True, include empirical PDF

        """

        # define colormap
        if cmap is not None:
            colormap = self.build_colormap(cmap, vmin)
        else:
            colormap = self.cmap

        # plot empirical pdf
        if empirical:
            ax.step(*self.epdf, where='post', color='r', linewidth=1)

        # plot individual component pdfs
        for i, pdf in enumerate(self.component_pdfs):
            color = colormap(self.component_to_label[i])
            if line:
                ax.plot(self.support, pdf, color=color, alpha=alpha, lw=1.)
            if fill:
                ax.fill_between(self.support, pdf, facecolors=color, alpha=alpha, linewidth=1., rasterized=True)

        # plot model pdf
        ax.plot(self.support, self.pdf, '--', c='k', lw=1)

        # format axis
        if ymax is None:
            maxima = self.pdf[argrelextrema(self.pdf, np.greater)]
            ymax = 2.5*np.product(maxima)**(1/maxima.size)

        ax.set_ylim(0, ymax)
        ax.set_xlim(self.model.lbound, self.model.ubound)
        ax.set_xlabel('Values', fontsize=8)
        ax.set_ylabel('Density', fontsize=8)

    @default_figure
    def plot_cdfs(self,
                  log=True,
                  cmap=plt.cm.Greys,
                  ax=None,
                  **kwargs):
        """
        Plot component cumulative distribution functions as stackplot.
        """

        def to_linear(support):
            """ Convert support to linear basis. """
            return np.exp(support)

        # log transform data
        support, esupport = self.support, self.esupport
        if not log:
            support, esupport = to_linear(support), to_linear(esupport)

        # get component CDFs
        component_cdfs = self.component_cdfs

        # plot weighted CDF for each component
        means = self.means
        values = self.values
        norm = Normalize(vmin=self.values.min(), vmax=self.values.max())
        order = np.argsort(means)
        colors = cmap(norm(means[order]))
        ax.stackplot(support, component_cdfs[order], colors=colors, **kwargs)

        # plot empirical CDF (data)
        ax.plot(esupport, self.ecdf, '-r', lw=1.)

        # plot mixture CDF
        ax.plot(support, component_cdfs.sum(axis=0), '--k', lw=1)


class BivariateMixtureVisualization:

    @property
    def support(self):
        """ Model support. """
        return self.model.supportx

    @property
    def esupport(self):
        """ Empirical support vector (sorted values). """
        return np.sort(self.values[:, 0])

    @property
    def pdf(self):
        """ Model PDF over support. """
        return self.model[0].pdf

    @property
    def component_pdfs(self):
        """ Weighted component PDFs over support. """
        return self.model[0].component_pdfs

    @property
    def component_cdfs(self):
        """ Returns weighted CDF of each component over support. """
        model = self.model[0]
        cdfs = np.vstack([x.cdf(self.support) for x in model.components])
        cdfs *= model.weights_.reshape(-1, 1)
        return cdfs

    @property
    def support_labels(self):
        """ Labels for support vector (over x margin). """
        margin = self.marginalize(0)
        return margin.classifier(self.support.reshape(-1, 1))

    @joint_figure
    def plot_bivariate_pdf(self, fig, bg='w', **kwargs):
        """ Plot bivariate PDF, with each cluster shaded by its label. """

        pdfs = self.model.component_pdfs

        # plot each bivariate pdf, colored by label
        norm = Normalize(pdfs.min(), pdfs.max())
        fig.ax_joint.set_facecolor(bg)
        for idx, pdf in enumerate(pdfs):
            ccm = build_transparent_cmap(self.label_colors[idx], bg=bg)
            fig.ax_joint.imshow(norm(pdf), cmap=ccm, extent=self.model.extent)

        # plot marginal pdfs
        self.model.plot_margin(0, ax=fig.ax_xmargin, component_color=self.label_colors)
        self.model.plot_margin(1, invert=True, ax=fig.ax_ymargin, component_color=self.label_colors)

        fig.ax_xmargin.set_xlim(self.model.lbound, self.model.ubound)
        fig.ax_ymargin.set_ylim(self.model.lbound, self.model.ubound)

        return fig

    @joint_figure
    def plot_bivariate_data(self, fig, bg='w', **kwargs):
        """ Plot bivariate data, with each sample shaded by its label. """

        # define label colors
        marker_labels = self.classifier(self.model.values)
        marker_colors = self.cmap(marker_labels)[:, :-1]

        # plot each bivariate pdf, colored by label
        fig.ax_joint.set_facecolor(bg)
        self.model.plot_data(ax=fig.ax_joint, c=marker_colors, **kwargs)
        fig.ax_joint.set_xlim(self.model.lbound, self.model.ubound)
        fig.ax_joint.set_ylim(self.model.lbound, self.model.ubound)
        fig.ax_joint.invert_yaxis()
        fig.ax_joint.set_yticks(fig.ax_joint.get_xticks())

        # plot marginal pdfs
        self.model.plot_margin(0, ax=fig.ax_xmargin, component_color=self.label_colors)
        self.model.plot_margin(1, invert=True, ax=fig.ax_ymargin, component_color=self.label_colors)
        fig.ax_xmargin.set_xlim(self.model.lbound, self.model.ubound)
        fig.ax_ymargin.set_ylim(self.model.lbound, self.model.ubound)

        return fig

    @joint_figure
    def plot_phase_space(self, fig, bg='w', **kwargs):
        """ Plot phase space, with each region shaded by its label. """

        labels = self.classifier(self.model.support)
        labels = labels.reshape(self.model.support_size)

        l, u = self.model.lbound, self.model.ubound
        extent = (l, u, u, l)
        fig.ax_joint.imshow(self.cmap(labels), extent=extent)

        # plot marginal pdfs
        self.model.plot_margin(0, ax=fig.ax_xmargin, component_color=self.label_colors)
        self.model.plot_margin(1, invert=True, ax=fig.ax_ymargin, component_color=self.label_colors)

        fig.ax_xmargin.set_xlim(self.model.lbound, self.model.ubound)
        fig.ax_ymargin.set_ylim(self.model.lbound, self.model.ubound)

        return fig



from os.path import join, exists
from os import mkdir
import pickle
from copy import deepcopy
import numpy as np
import matplotlib.pyplot as plt

from ...utilities import IO
from ..mixtures import UnivariateMixture, BivariateMixture

from .classifiers import Classifier, ClassifierIO
from .visualization import MixtureVisualization, BivariateMixtureVisualization


class MixtureModelIO(ClassifierIO):
    """
    Methods for saving and loading classifier objects.
    """

    def save(self, dirpath, data=False, image=True, extension=None, **kwargs):
        """
        Save classifier to specified path.

        Args:

            dirpath (str) - directory in which classifier is to be saved

            data (bool) - if True, save training data

            image (bool) - if True, save labeled histogram image

            extension (str) - directory name extension

            kwargs: keyword arguments for image rendering

        """

        # instantiate Classifier
        path = super().save(dirpath, data, image, extension, **kwargs)

        # save model (temporarily remove values)
        if self.model is not None:
            #self.model.values = None
            with open(join(path, 'model.pkl'), 'wb') as file:
                pickle.dump(self.model, file)

        return path

    @classmethod
    def load(cls, path):
        """
        Load classifier from file.

        Args:

            path (str) - path to classifier directory

        Returns:

            classifier (Classifier derivative)

        """
        io = IO()

        values_path = join(path, 'values.npy')
        if exists(values_path):
            values = io.read_npy(values_path)
        else:
            values = None

        parameters = io.read_json(join(path, 'parameters.json'))

        # load model
        with open(join(path, 'model.pkl'), 'rb') as file:
            model = pickle.load(file)

        if values is not None:
            model.values = np.log(values)

        return cls(values, model=model, **parameters)


class UnivariateMixtureClassifier(MixtureModelIO,
                                  Classifier,
                                  MixtureVisualization):
    """
    Univariate mixed log-normal model classifier.

    Attributes:

        model (mixtures.UnivariateMixture) - frozen univariate mixture model

        num_components (int) - number of mixture components

        classifier (vectorized func) - maps values to labels

        labels (np.ndarray[int]) - predicted labels

    Inherited attributes:

        values (np.ndarray[float]) - basis for clustering

        num_labels (int) - number of output labels

        log (bool) - indicates whether clustering performed on log values

        cmap (matplotlib.colors.ColorMap) - colormap for labels

        parameters (dict) - {param name: param value} pairs

    """

    def __init__(self, values,
                 num_components=3,
                 num_labels=3,
                 fit_kw={},
                 model=None,
                 **kwargs):
        """
        Fit a univariate mixture model classifier to an array of values.

        Args:

            values (np.ndarray[float]) - basis for clustering (not log-transformed)

            num_components (int) - number of mixture components

            num_labels (int) - number of class labels

            fit_kw (dict) - keyword arguments for fitting mixture model

            model (mixtures.UnivariateMixture) - pre-fitted model

        Keyword arguments:

            attribute (str or list) - attribute(s) on which to cluster

            cmap (matplotlib.colors.ColorMap) - colormap for class_id

        """

        # instantiate classifier (remove redundant log parameter)
        if 'log' in kwargs.keys():
            _ = kwargs.pop('log')
        super().__init__(values, num_labels=num_labels, log=True, **kwargs)
        self.parameters['num_components'] = num_components
        self.parameters['fit_kw'] = fit_kw

        # fit model
        if model is None:
            model = self.fit(self.values, num_components, **fit_kw)
        self.model = model

        # build classifier and posterior
        self.classifier = self.build_classifier()
        self.posterior = self.build_posterior()

        # assign labels
        if values is not None:
            self.labels = self.classifier(self.values)

    @property
    def num_components(self):
        """ Number of model components. """
        return self.model.n_components

    @property
    def means(self):
        """ Mean of each component. """
        return self.model.means

    @staticmethod
    def fit(values, num_components=3, **kwargs):
        """
        Fit univariate gaussian mixture model.

        Args:

            values (np.ndarray[float]) - 1D array of log-transformed values

            num_components (int) - number of model components

            kwargs: keyword arguments for fitting

        Returns:

            model (mixtures.UnivariateMixture)

        """
        return UnivariateMixture.from_logsample(values,
                                                num_components,
                                                **kwargs)

    def predict(self, values):
        """ Predict which component each of <values> belongs to. """
        return self.model.predict(values)

    def predict_proba(self, values):
        """
        Predict the posterior probability with which each of <values> belongs to each component.
        """
        return self.model.predict_proba(values)

    def build_posterior(self):
        """
        Build function that returns the posterior probability of each label given a series of values.
        """

        def posterior(values):
            """ Returns probabilities of each label for <values>.  """

            # evaluate posterior probability of each label for each value
            p = self.model.predict_proba(values)
            _posterior = [p[:,i].sum(axis=1) for i in self.component_groups]
            _posterior = np.vstack(_posterior).T

            # fix label probabilities for points outside the support bounds
            below = values[:, 0] < self.model.lbound
            above = values[:, 0] > self.model.ubound
            for rows, col in zip((below.ravel(), above.ravel()), [0, -1]):
                if rows.sum() == 0:
                    continue
                adjust = np.zeros((rows.sum(), self.num_labels), dtype=float)
                adjust[:, col] = 1.
                _posterior[rows] = adjust

            return _posterior

        return posterior

    def evaluate_posterior(self, data):
        """ Returns posterior across components for <data>. """
        x =  data[self.attribute].values
        if self.log:
            x = np.log(x)
        return self.posterior(x)

    def build_classifier(self):
        """
        Build function that returns the most probable label for each of a series of values.
        """

        def classifier(values):
            """ Returns <label> for <values> by maximizing posterior.  """
            return self.posterior(values).argmax(axis=1)

        return classifier


class BivariateMixtureClassifier(BivariateMixtureVisualization,
                                 UnivariateMixtureClassifier):
    """
    Bivariate mixed log-normal model classifier.

    Attributes:

        model (mixtures.BivariateMixture) - frozen bivariate mixture model

    Inherited attributes:

        values (np.ndarray[float]) - basis for clustering

        attribute (list) - attributes on which to cluster

        num_labels (int) - number of labels

        num_components (int) - number of mixture components

        classifier (vectorized func) - maps values to labels

        labels (np.ndarray[int]) - predicted labels

        log (bool) - indicates whether clustering performed on log values

        cmap (matplotlib.colors.ColorMap) - colormap for labels

        parameters (dict) - {param name: param value} pairs

    """

    def __getitem__(self, margin):
        """ Returns UnivariateMixtureClassifier for specified margin. """
        return self.marginalize(margin)

    def marginalize(self, margin):
        """ Returns UnivariateMixtureClassifier for specified margin. """

        # assemble marginalized properties
        values = self._values[:, [margin]]
        model = self.model[margin]

        # duplicate parameters
        parameters = deepcopy(self.parameters)
        parameters['attribute'] = self.attribute[0]
        _ = parameters.pop('log')

        return UnivariateMixtureClassifier(values, model=model, **parameters)

    @staticmethod
    def fit(values, num_components=3, **kwargs):
        """
        Fit univariate gaussian mixture model.

        Args:

            values (np.ndarray[float]) - 1D array of log-transformed values

            num_components (int) - number of model components

            kwargs: keyword arguments for fitting

        Returns:

            model (mixtures.BivariateMixture)

        """
        return BivariateMixture.from_logsample(values, num_components, **kwargs)



from unittest import TestCase
from unittest.mock import patch

from os.path import join
from shutil import move, rmtree
from flyqma.data import Experiment, Stack


class TestPaths(TestCase):
    """ Initialize paths to test data. """
    bit_depth = 12
    fixtures_path = 'flyqma/tests/fixtures'
    experiment_path = join(fixtures_path, 'experiment')
    stack_path = join(experiment_path, 'disc')
    #silhouette_path = join(self.fixtures_path, 'disc.silhouette')
    tif_path = join(fixtures_path, 'disc.tif')
    roi_mask_path = join(fixtures_path, 'roi_mask.npy')
    segmentation_mask_path = join(fixtures_path, 'segmentation_mask.npy')


class Test01_Instantiation(TestPaths):
    """
    Tests for instantiating Experiment and Stack instances in Fly-QMA.
    """

    def test00_instantiate_stack_from_path(self):
        """ Instantiate Stack from stack path.  """
        with patch('builtins.input', side_effect=['yes', 12]):
            stack = Stack(self.stack_path, bit_depth=self.bit_depth)
        self.assertTrue(isinstance(stack, Stack))

    #def test01_instantiate_stack_from_silhouette(self):
        """ Instantiate Stack from Silhouette file.  (not yet implemented) """
        #stack = Stack.from_silhouette(self.silhouette_path)
        #self.assertTrue(isinstance(stack, Stack))
        #self.assertTrue(True)

    def test02_instantiate_experiment_from_path(self):
        """ Instantiate experiment from collection of stack directories.  """
        with patch('builtins.input', side_effect=['yes', 12]):
            experiment = Experiment(self.experiment_path)
        self.assertTrue(isinstance(experiment, Experiment))

    def test03_instantiate_stack_from_tif(self):
        """ Instantiate Stack from tif image.  """
        with patch('builtins.input', side_effect=['yes', 12]):
            stack = Stack.from_tif(self.tif_path, bit_depth=self.bit_depth)
        self.assertTrue(isinstance(stack, Stack))

        # restore original tif and remove stack directory
        move(stack.tif_path, join(stack.path, '..'))
        stack.restore_directory()
        rmtree(stack.path)



from os.path import join
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from matplotlib.colors import ListedColormap

from .settings import *


class Figure:
    """
    Base class for figures providing some common methods.

    Attributes:

        name (str) - figure name

        directory (str) - default path for saving figure

        fig (matplotlib.figure.Figure)

        axes (matplotlib.axes.AxesSubplots)

    """

    def __init__(self, name='unnamed', directory='../graphics'):
        self.name = name
        self.directory = directory
        self.fig = None

    @staticmethod
    def create_figure(figsize=(3, 3)):
        """
        Create blank figure.

        Args:

            figsize (tuple) - figure dimensions
        """
        fig = plt.figure(figsize=figsize)
        return fig

    def add_axes(self, nrows=1, ncols=1):
        """
        Add axes to figure.

        Args:

            nrows, ncols (int) - number of rows and columns

        """
        self.axes = self.fig.subplots(nrows=nrows, ncols=ncols)

    def save(self, **kwargs):
        """
        Save figure to file.

        Keyword Arguments:

            fmt (str) - file format, eg 'pdf'

            dpi (int) - resolution

            transparent (bool) - if True, remove background

            rasterized (bool) - if True, rasterize figure data

            addtl kwargs: keyword arguments for plt.savefig

        """

        self._save(self.fig, self.name, self.directory, **kwargs)

    @staticmethod
    def _save(fig,
             name,
             dirpath='./',
             fmt='pdf',
             dpi=300,
             transparent=True,
             rasterized=True,
             **kwargs):
        """
        Save figure to file.

        Args:

            fig (matplotlib.figures.Figure) - figure to be saved

            name (str) - file name without format extension

            dirpath (str) - directory in which to save file

            fmt (str) - file format, eg 'pdf'

            dpi (int) - resolution

            transparent (bool) - if True, remove background

            rasterized (bool) - if True, rasterize figure data

            kwargs: keyword arguments for plt.savefig

        """
        path = join(dirpath, name+'.{}'.format(fmt))
        kw = dict(dpi=dpi, transparent=transparent, rasterized=rasterized)
        fig.savefig(path, format=fmt, **kwargs)

    def _add_markers(self, x, y, c, **kwargs):
        """
        Add markers to axis.

        Args:

            x, y (array like) - marker x and y positions

            c (array like) - marker colors

            kwargs: keyword arguments for matplotlib.pyplot.scatter

        """

        if len(self.fig.axes) == 0:
            ax = self.fig.subplots()
        ax = self.fig.axes[0]

        # add markers to plot
        ax.scatter(x, y, c=c, **kwargs)

    def format(self, **kwargs):
        """ Format all figure axes. """
        for ax in self.fig.axes:
            self.format_axis(ax, **kwargs)

    def format_axis(self, ax):
        """ Format individual axis. """
        pass


class CellSelection(Figure):
    """
    Visualize cell selection by overlaying cell position markers on top of an image of a single RGB layer.

    Inherited attributes:

        name (str) - figure name ('selection')

        directory (str) - default path for saving figure

        fig (matplotlib.figure.Figure)

        axes (matplotlib.axes.AxesSubplots)

    """

    def __init__(self, layer, data, channel='r', **kwargs):
        """
        Instantiate cell selection figure.

        Args:

            layer (Layer) - RGB image layer

            data (pd.DataFrame) - selected cell measurement data

            channel (str) - color channel to be added

            kwargs: keyword arguments for render

        """
        Figure.__init__(self, name='selection')
        self.render(layer, data, **kwargs)

    def render(self, layer, data, channel='r', figsize=(3, 3)):
        """
        Render figure.

        Args:

            layer (Layer) - RGB image layer

            data (pd.DataFrame) - selected cell measurement data

            channel (str) - color channel to be added

            figsize (tuple) - figure dimensions

        """

        # create figure
        self.fig = self.create_figure(figsize)
        self.add_axes()

        # add image
        self.add_image(layer, channel=channel)

        # add cell position markers
        self.add_markers(data)

    def add_image(self, layer, channel='r'):
        """
        Add scalar image to figure.

        Args:

            layer (Layer) - RGB image layer

            channel (str) - color channel to be added

        """
        _ = layer.get_channel(channel).show(ax=ax, segments=False, cmap=None)
        _ = ax.axis('off')

    def add_markers(self, data, color_by='genotype', xykey=None, **kwargs):
        """
        Add cell position markers to axis.

        Args:

            data (pd.DataFrame) - selected cell measurement data

            color_by (str) - cell measurement attribute used to color markers

            xykey (list) - attribute keys for cell x/y positions

            kwargs: keyword arguments for markers

        """

        if xykey is None:
            xykey = ['centroid_x', 'centroid_y']

        # get cell coordinates and color vector
        x, y = data[xykey].values.T

        # get color vector and colormap
        c = data[color_by]
        cmap = ListedColormap(['y', 'c', 'm'], 'indexed', 3)

        # add markers to plot
        self._add_markers(x, y, c, cmap=cmap, vmin=0, vmax=2, **kwargs)


class Scatterplot(Figure):
    """
    Scatter points in XY plane.

    Attributes:

        xvar, yvar (str) - cell measurement features to be scattered

    Inherited attributes:

        name (str) - figure name

        directory (str) - default path for saving figure

        fig (matplotlib.figure.Figure)

        axes (matplotlib.axes.AxesSubplots)

    """

    def __init__(self, data, xvar, yvar, name, **kwargs):
        """
        Instantiate scatter plot.

        Args:

            data (pd.DataFrame) - selected cell measurement data

            xvar, yvar (str) - cell measurement features to be scattered

            name (str) - figure name

            kwargs: keyword arguments for

        """
        Figure.__init__(self, name=name)
        self.xvar, self.yvar = xvar, yvar
        self.render(data, **kwargs)

    def render(self, data, figsize=(2, 2)):
        """
        Render figure.

        Args:

            data (pd.DataFrame) - selected cell measurement data

            figsize (tuple) - figure dimensions

        """

        # create figure
        self.fig = self.create_figure(figsize)
        self.add_axes()

        # add data
        self._add_markers(data[self.xvar], data[self.yvar], c='k', s=1)

        # format axes
        self.format()

    def format_axis(self, ax):
        """
        Format axis.

        Args:

            ax (matplotlib.axes.AxesSubplot)

        """
        _ = ax.spines['top'].set_visible(False)
        _ = ax.spines['right'].set_visible(False)
        _ = ax.set_xlabel(self.x)
        _ = ax.set_ylabel(self.y)


class BackgroundCorrelation(Scatterplot):
    """
    Plot correlated expression between red and green fluorescence channels.

    Inherited attributes:

        xvar, yvar (str) - cell measurement features to be scattered

        name (str) - figure name

        directory (str) - default path for saving figure

        fig (matplotlib.figure.Figure)

        axes (matplotlib.axes.AxesSubplots)

    """

    def __init__(self, data, name, figsize=(2, 2)):
        """
        Instantiate background correlation plot.

        Args:

            data (pd.DataFrame) - selected cell measurement data

            name (str) - figure name

            figsize (tuple) - figure size

        """
        Scatterplot.__init__(self, data, 'r', 'g', name, figsize=figsize)

    def format_axis(self, ax):
        """
        Format axis.

        Args:

            ax (matplotlib.axes.AxesSubplot)

        """
        _ = ax.spines['top'].set_visible(False)
        _ = ax.spines['right'].set_visible(False)
        _ = ax.set_xticks(np.arange(0, .95, .2))
        _ = ax.set_yticks(np.arange(0, .95, .2))
        _ = ax.set_xlabel('Nuclear RFP level')
        _ = ax.set_ylabel('Nuclear GFP level')



from os.path import join
import gc
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from ..utilities import IO
from ..visualization import *

from .models import GLM
from .resampling import resample_uniformly
from .background import BackgroundExtraction
from .visualization import CorrectionVisualization
from .visualization import LayerCorrectionVisualization


class Correction(GLM, CorrectionVisualization):
    """
    Linear correction for background correlation between fluorescence channels within an individual layer.

    Attributes:

        xt, yt (np.ndarray[float]) - foreground measurements

        xraw, yraw (np.ndarray[float]) - raw background pixel intensities

        x, y (np.ndarray[float]) - resampled background pixel intensities

    """

    def __init__(self, xt, yt, bg_x, bg_y,
                 remove_zeros=False,
                 resample=True,
                 resample_size=None,
                 resample_cutoff=None,
                 store_pixels=True,
                 **fit_kw):
        """
        Instantiate bleedthrough correction for an RGB image layer.

        Args:

            data (pd.DataFrame) - measurement data

            xt, yt (np.ndarray[float]) - foreground measurements

            bg_x, bg_y (np.ndarray[float]) - background pixel intensities

            remove_zeros (bool) - if True, remove all zero-valued pixels.

            resample (bool) - if True, uniformly resample pixels in X

            resample_size (int) - number of uniform samples

            resample_cutoff (int) - upper bound for samples (quantile, 0-100)

            store_pixels (bool) - if True, store raw background pixels

        """

        # store data
        self.xt = xt
        self.yt = yt

        self.store_pixels = store_pixels
        if store_pixels:
            self.xraw = bg_x
            self.yraw = bg_y

        # store parameters
        self.remove_zeros = remove_zeros
        self.resample = resample
        self.resample_size = resample_size
        self.resample_cutoff = resample_cutoff

        # remove zero-valued pixels
        if remove_zeros:
            bg_x, bg_y = self._remove_zeros(bg_x, bg_y)

        # resample uniformly in X
        if resample:
            bg_x, bg_y = resample_uniformly(bg_x, bg_y, resample_size, resample_cutoff)

        # fit line to background pixels
        super().__init__(bg_x, bg_y, **fit_kw)

        # apply correction to measurements (internally)
        self.correct_measurements()

        # instantiate container for figures
        self.figs = {}

    def correct_measurements(self):
        """ Apply correction to measurements. """

        # store measurement values (test data)
        self.xtdomain = np.linspace(0, self.xt.max(), 10)

        # store model prediction and corrected measurements
        self.ytp = self.predict(self.xt)
        self.ytc = self.yt - self.ytp

    @staticmethod
    def _remove_zeros(x, y):
        """ Remove pixels with zero values in either channel. """
        nonzero_mask = np.logical_and(x!=0, y!=0)
        return x[nonzero_mask], y[nonzero_mask]


class LayerCorrection(Correction, LayerCorrectionVisualization):
    """
    Linear correction for background correlation between fluorescence channels within an individual layer.

    Attributes:

        layer (Layer) - layer RGB image

    Inherited attributes:

        xt, yt (np.ndarray[float]) - foreground measurements

        xraw, yraw (np.ndarray[float]) - raw background pixel intensities

        x, y (np.ndarray[float]) - resampled background pixel intensities

    Parameters:

        xvar (str) - name of independent variable attribute in measurement data

        yvar (str) - name of dependent variable attribute in measurement data

        niters (int) - number of binary dilations applied to foreground mask

        remove_zeros (bool) - if True, remove all zero-valued pixels.

        resample (bool) - if True, uniformly resample pixels in X

        resample_size (int) - number of uniform samples

        resample_cutoff (int) - upper bound for samples (quantile, 0 to 100)

    """

    def __init__(self, layer, xvar, yvar,
                 niters=50,
                 remove_zeros=False,
                 resample=True,
                 resample_size=None,
                 resample_cutoff=None,
                 store_pixels=False,
                 **fit_kw):
        """
        Instantiate bleedthrough correction for an RGB image layer.

        Args:

            layer (Layer) - layer RGB image

            xvar (int) - independent color channel

            yvar (int) - dependent color channel

            niters (int) - number of binary dilations applied to foreground

            remove_zeros (bool) - if True, remove all zero-valued pixels.

            resample (bool) - if True, uniformly resample pixels in X

            resample_size (int) - number of uniform samples

            resample_cutoff (int) - upper bound for samples (quantile, 0-100)

            store_pixels (bool) - if True, store raw background pixels

        """

        # store layer
        self.layer = layer
        self.xvar = xvar
        self.yvar = yvar
        self.niters = niters

        # get foreground measurements
        xt = layer.data[self.xkey].values
        yt = layer.data[self.ykey].values

        # extract X and Y pixels from background
        bg_x, bg_y = self.extract_background()

        # instantiate correction
        super().__init__(xt, yt, bg_x, bg_y,
                         remove_zeros=remove_zeros,
                         resample=resample,
                         resample_size=resample_size,
                         resample_cutoff=resample_cutoff,
                         store_pixels=store_pixels,
                         **fit_kw)

    @property
    def xkey(self):
        """ DataFrame key for independent channel. """
        return 'ch{:d}'.format(self.xvar)

    @property
    def ykey(self):
        """ DataFrame key for dependent channel. """
        return 'ch{:d}'.format(self.yvar)

    def extract_background(self):
        """ Returns raw background pixels. """
        bg_ext = BackgroundExtraction(self.layer, self.niters)
        bg_x = bg_ext.extract_pixels(self.xvar)
        bg_y = bg_ext.extract_pixels(self.yvar)
        return bg_x, bg_y

    @classmethod
    def load(cls, layer):
        """
        Load linear model from file.

        Args:

            path (str) - path to correction directory

        Returns:

            correction (LayerCorrection)

        """

        path = layer.subdirs['correction']

        # load data
        io = IO()
        data = io.read_json(join(path, 'data.json'))

        return LayerCorrection(layer, **data['params'])

    def save(self, images=True):
        """
        Save linear model and corrected levels.

        Args:

            images (bool) - if True, save model fit and corrected measurement figs

        """

        # add subdirectory to layer
        self.layer.make_subdir('correction')
        path = self.layer.subdirs['correction']

        # instantiate IO
        io = IO()

        # save metadata to json
        params = dict(
                    xvar=self.xvar,
                    yvar=self.yvar,
                    niters=self.niters,
                    remove_zeros=self.remove_zeros,
                    resample=self.resample,
                    resample_size=self.resample_size,
                    resample_cutoff=self.resample_cutoff)

        data = dict(mode=self.__class__.__name__,
                    params=params,
                    coefficients=self.model.params.tolist())

        # write metadata to file
        io.write_json(join(path, 'data.json'), data)

        # update measurements
        self.layer.apply_correction(self.layer.data)
        self.layer.save_processed_data()

        # save figures
        if images:
            self.show_fit()
            self.show_correction()
            self.save_figs()

    def save_figs(self, dpi=100, fmt='png'):
        """
        Save all figures.

        Args:

            dpi (int) - resolution

            fmt (str) - image format

        """

        # get correction path
        dirpath = self.layer.subdirs['correction']

        # keyword arguments for savefig
        kw = dict(dpi=dpi, format=fmt, transparent=True, rasterized=True)

        for name, fig in self.figs.items():

            # save figure
            path = join(dirpath, name+'.png')
            fig.savefig(path, **kw)

            # close figure
            fig.clf()
            plt.close(fig)
            gc.collect()



from math import floor
import numpy as np
import scipy.stats as st
from scipy.signal import savgol_filter
import matplotlib.pyplot as plt
import warnings


def savgol(x, window_size=100, polyorder=1):
    """
    Perform Savitzky-Golay filtration of 1-D array.

    Args:

        x (np.ndarray) - ordered samples

        window_size (int) - filter size

        polyorder (int) - polynomial order

    Returns:

        trend (np.ndarray) - smoothed values

    """

    # window size must be odd
    if window_size % 2 == 0:
        window_size += 1
    window_size = int(window_size)

    with warnings.catch_warnings():

        # filter multidimensional indexing warning (fixed in scipy v1.2.0)
        warnings.filterwarnings('ignore', category=FutureWarning)

        # filter scipy gelsd warning
        warnings.filterwarnings('ignore', message='^internal gelsd')

        # apply savgol filter
        trend = savgol_filter(x, window_size, polyorder=polyorder)

    return trend


def get_running_mean(x, window_size=100):
    """
    Returns running mean for a 1D vector. This is the fastest implementation, but is limited to one-dimensional arrays and doesn't permit interval specification.

    Args:

        x (np.ndarray) - ordered samples, length N

        window_size (int) - size of window, W

    Returns:

        means (np.ndarray) - moving average of x

    """
    cumsum = np.cumsum(np.insert(x, 0, 0))
    return (cumsum[window_size:] - cumsum[:-window_size]) / window_size


def get_rolling_window(x, window_size=100, resolution=1):
    """
    Return array slices within a rolling window.

    Args:

        x (np.ndarray) - ordered samples, length N

        window_size (int) - size of window, W

        resolution (int) - sampling interval

    Returns:

        windows (np.ndarray) - sampled values, N/resolution x W

    """

    if resolution is None:
        resolution = window_size

    shape = x.shape[:-1] + (floor((x.shape[-1] - window_size + 1)/resolution), window_size)
    strides = x.strides[:-1] + (x.strides[-1]*resolution,) + (x.strides[-1],)
    return np.lib.stride_tricks.as_strided(x, shape=shape, strides=strides)


def get_rolling_mean(x, **kw):
    """
    Compute rolling mean. This implementation permits flexible sampling intervals and multi-dimensional time series, but is slower than get_running_mean for 1D time series.

    Args:

        x (np.ndarray) - ordered samples, length N

        kw: arguments for window specification

    Returns:

        means (np.ndarray) - moving average of x, N/resolution x 1

    """
    return get_rolling_window(x, **kw).mean(axis=-1)


def get_binned_mean(x, window_size=100):
    """
    Returns mean values within non-overlapping sequential windows.

    Args:

        x (np.ndarray) - ordered samples, length N

        window_size (int) - size of window, W

    Returns:

        means (np.ndarray) - bin means, N/W x 1

    """
    return get_rolling_mean(x, window_size=window_size, resolution=window_size)


def apply_custom_roller(func, x, **kwargs):
    """
    Apply function to rolling window.

    Args:

        func (function) - function applied to each window, returns 1 x N_out

        x (np.ndarray) - ordered samples, length N

        kwargs: keyword arguments for window specification

    Returns:

        fx (np.ndarray) - function output for each window, N/resolution x N_out

    """
    windows = get_rolling_window(x, **kwargs)
    return np.apply_along_axis(func, axis=-1, arr=windows)


def subsample(x, frac=1):
    """
    Subsample array with replacement.

    Args:

        x (np.ndarray) - ordered samples, length N

        frac (float) - sample size (fraction of array)

    Returns:

        sample (np.ndarray) - subsampled values

    """
    return x[np.random.randint(0, len(x), size=floor(frac*len(x)))]


def bootstrap(x, func=np.mean, confidence=95, N=1000):
    """
    Returns point estimate obtained by parametric bootstrap.

    Args:

        x (np.ndarray) - ordered samples, length N

        func (function) - function applied to each bootstrap sample

        confidence (float) - confidence interval, between 0 and 100

        N (int) - number of bootstrap samples

    Returns:

        interval (np.ndarray) - confidence interval bounds

    """
    point_estimates = [func(subsample(x)) for _ in range(N)]
    q = (((100-confidence)/2), (100+confidence)/2)
    return np.percentile(point_estimates, q=q)


def get_rolling_mean_interval(x,
                              window_size=100,
                              resolution=1,
                              confidence=95,
                              nbootstraps=1000):
    """
    Evaluate confidence interval for moving average of ordered values.

    Args:

        x (np.ndarray) - ordered samples, length N

        window_size (int) - size of window, W

        resolution (int) - sampling interval

        confidence (float) - confidence interval, between 0 and 100

        nbootstraps (int) - number of bootstrap samples

    Returns:

        interval (np.ndarray) - confidence interval bounds, N/resolution x 2

    """

    # define bootstrapping operation
    f = lambda s: bootstrap(s, np.mean, confidence=confidence, N=nbootstraps)

    # apply to moving windows
    kw = dict(window_size=window_size, resolution=resolution)
    interval = apply_custom_roller(f, x, **kw)

    return interval


def get_rolling_gaussian(x, window_size=100, resolution=10):
    """
    Returns gaussian fit within sliding window.

    Args:

        x (np.ndarray) - ordered samples

        window_size (int) - size of window

        resolution (int) - sampling interval

    Returns:

        model (scipy.stats.norm)

    """

    # assemble sliding windows
    kw = dict(window_size=window_size, resolution=resolution)
    windows = get_rolling_window(x, **kw)

    # fit gaussians
    loc = windows.mean(axis=-1)
    scale = np.sqrt(windows.var(axis=-1))
    model = st.norm(loc=loc, scale=scale)

    return model


def detrend_signal(x, window_size=99, order=1):
    """
    Detrend and scale fluctuations using first-order univariate spline.

    Args:

        x (np array) -ordered samples

        window_size (int) - size of interpolation window for lowpass filter

        order (int) - spline order

    Returns:

        residuals (np array) - detrended residuals

        trend (np array) - spline fit to signal

    """

    # use odd window size
    if window_size % 2 == 0:
        window_size += 1
    window_size = int(window_size)

    # evaluate trend
    trend = savgol(x, window_size=window_size, polyorder=order)

    # evaluate residuals (catch warnings for invalid values)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        residuals = x - trend

    return residuals, trend


def smooth(x, window):
    """ Returns smoothed moving average of <x> within <window>. """
    kw = dict(window_size=window, resolution=int(window/5))
    return savgol(get_rolling_mean(x, **kw), window)


def plot_mean(ax, x, y,
              label=None,
              ma_type='sliding',
              window_size=100,
              resolution=1,
              line_color='k',
              line_width=1,
              line_alpha=1,
              linestyle=None,
              markersize=2,
              smooth=False,
              **kw):
    """
    Plot moving average.

    Args:

        x, y (array like) - timeseries data

        ax (matplotlib.axes.AxesSubplot) - axis which to which line is added

        label (str) - data label

        ma_type (str) - type of average used, either sliding, binned, or savgol

        window_size (int) - size of window

        resolution (int) - sampling interval

        line_color, line_width, line_alpha, linestyle - formatting parameters

        smooth (bool) - if True, apply secondary savgol filter

    Returns:

        line (matplotlib.lines.Line2D)

    """

    # get moving average (skip first point to avoid outliers)
    if ma_type == 'savgol':
        x_av = x[1:]
        y_av = savgol(y, window_size=window_size, polyorder=1)[1:]
    else:
        if ma_type == 'binned':
            resolution = window_size
        x_av = get_rolling_mean(x, window_size=window_size, resolution=resolution)
        y_av = get_rolling_mean(y, window_size=window_size, resolution=resolution)

    # get line and dashstyles
    if linestyle == None:
        linestyle = 'solid'
    dashstyles = {'solid': (None, None), 'dashed': (2.0, 2.0)}
    dashstyle = dashstyles[linestyle]

    # apply secondary smoothing (for visualization)
    if smooth:
        sw = int(window_size/3)
        if sw > 1:
            y_av = savgol(y_av, window_size=sw, polyorder=1)

    # plot line
    line = ax.plot(x_av, y_av,
                   linestyle=linestyle, dashes=dashstyle,
                   lw=line_width, color=line_color, alpha=line_alpha,
                   label=label, markersize=markersize)

    return line


def plot_mean_interval(ax, x, y,
                       ma_type='sliding',
                       window_size=100,
                       resolution=10,
                       nbootstraps=1000,
                       confidence=95,
                       color='grey',
                       alpha=0.25,
                       error_bars=False,
                       lw=0.):
    """
    Adds confidence interval for line average (sliding window or binned) to existing axes.

    Args:

        x, y (array like) - data

        ax (axes) - axis which to which line is added

        ma_type (str) - type of average used, either 'sliding' or 'binned'

        window_size (int) - size of sliding window or bin (num of cells)

        interval_resolution (int) - sampling resolution for confidence interval

        nbootstraps (int) - number of bootstraps

        confidence (float) - confidence interval, between 0 and 100

        color, alpha - formatting parameters

    """

    if ma_type == 'binned':
        resolution = window_size
    x_av = get_rolling_mean(x, window_size=window_size, resolution=resolution)
    y_av = get_rolling_mean(y, window_size=window_size, resolution=resolution)
    interval = get_rolling_mean_interval(y, window_size=window_size, resolution=resolution, nbootstraps=nbootstraps, confidence=confidence)
    y_lower, y_upper = interval.T
    _ = ax.fill_between(x_av, y_lower, y_upper,
                        color=color, alpha=alpha, lw=lw)

    if error_bars == True:
        ax.errorbar(x_av, y_av, yerr=[y_av-y_lower, y_upper-y_av], fmt='-o', color=color)




from clones.validation.arguments import SweepArguments
from clones.validation.sweep import SweepBenchmark


# ======================== PARSE SCRIPT ARGUMENTS =============================

args = SweepArguments(description='Benchmark sweep options.')
sweep_path = args['path']
min_ambiguity = args['min_ambiguity']
max_ambiguity = args['max_ambiguity']
num_ambiguities = args['num_ambiguities']
num_replicates = args['num_replicates']

# ============================= RUN SCRIPT ====================================

# instantiate SweepBenchmark object
sweep = SweepBenchmark(
    sweep_path,
    min_ambiguity=min_ambiguity,
    max_ambiguity=max_ambiguity,
    num_ambiguities=num_ambiguities,
    num_replicates=num_replicates)

# build SweepBenchmark
sweep.build(
    logratio=args['logratio'],
    train_globally=args['train_globally'],
    walltime=args['walltime'],
    cores=args['cores'],
    memory=args['memory'],
    allocation=args['allocation'],
    attribute='clonal_marker')



from distutils.util import strtobool


class UserPrompts:
    """
    Event handling for user prompts.
    """

    @classmethod
    def boolean_prompt(cls, question, num_attempts=0, hint=' [y/n] '):
        """ Prompt user for boolean response to <question>. """

        response = str(input(question + hint))

        try:
            if strtobool(response):
                return True
            else:
                return False
        except:
            print('Response not recognized, please try again.')
            if num_attempts >= 3:
                return False
            return cls.boolean_prompt(question, num_attempts+1, hint)

    @classmethod
    def integer_prompt(cls, question, num_attempts=0, hint=' [0-16] '):
        """ Prompt user for integer response to <question>. """

        response = str(input(question + hint))

        try:
            return int(response)

        except:
            print('Integer value required, please try again.')
            if num_attempts >= 3:
                return None
            return cls.integer_prompt(question, num_attempts+1, hint)



import numpy as np

try:
    import infomap
except ImportError:
    raise UserWarning('No infomap package found. In order to use the community-based annotation scheme you must install it via PyPI.')


class InfoMap:
    """
    Object for performing infomap flow-based community detection.

    Attributes:

        infomap (infomap.Infomap) - infomap object

        node_to_module (dict) - {node: module} pairs

        classifier (vectorized func) - maps nodes to modules

        aggregator (CommunityAggregator)

    """

    def __init__(self, edges, **kwargs):
        """
        Instantiate infomap community detection. Two-level community detection is used by default.

        Args:

            edges (list) - (i, j, weight) tuple for each edge

            kwargs: keyword arguments for build_network method, including:

                twolevel (bool) - if True, perform two-level clustering

                N (int) - number of trials

        """

        self.infomap = self.build_network(edges, **kwargs)
        self.run()
        node_to_module, classifier = self.build_classifier()
        self.node_to_module = node_to_module
        self.classifier = classifier
        self.aggregator = CommunityAggregator(self.infomap)

    def __call__(self, x, level=None):
        """ Returns predicted class labels for values. """
        return self.aggregator(self.classifier(x), level)

    @property
    def max_depth(self):
        """ Maximum tree depth. """
        return self.infomap.maxTreeDepth()

    @staticmethod
    def build_network(edges, twolevel=False, N=25):
        """
        Compile InfoMap object from graph edges.

        Args:

            twolevel (bool) - if True, perform two-level clustering

            N (int) - number of trials

        """

        # define arguments
        args = '--undirected --silent -N {:d}'.format(N)
        if twolevel:
            args = '--two-level ' + args

        # instantiate infomap
        infomap_obj = infomap.Infomap(args)
        network = infomap_obj.network()

        # add edges
        _ = [network.addLink(*e) for e in edges]

        return infomap_obj

    def run(self, report=False):
        """
        Run infomap community detection.

        Args:

            report (bool) - if True, print number of modules found

        """
        self.infomap.run()
        if report:
            print("Found {:d} modules.".format(self.infomap.numTopModules()))

    def build_classifier(self):
        """
        Construct node to module classifier.

        Returns:

            node_to_module (dict) - {node: module} pairs

            classifier (vectorized func) - maps nodes to modules

        """
        node_to_module = {}
        for node in self.infomap.iterLeafNodes():
            node_to_module[node.physicalId] = node.moduleIndex()
        return node_to_module, np.vectorize(node_to_module.get)


class CommunityAggregator:
    """ Tool for hierarchical aggregation of communities. """

    def __init__(self, infomap):
        self.infomap = infomap
        self.max_depth = self.infomap.maxTreeDepth()

    def __getitem__(self, depth):
        """ Returns dictionary mapping low level modules to higher modules. """
        return self.group_modules(depth)

    def __call__(self, modules, level=None):
        """ Returns labels for modules cut to <level>. """
        return self.group(modules, level)

    @property
    def module_to_paths(self):
        return {m.moduleIndex(): m.path() for m in self.infomap.iterModules() if m.isLeafModule()}

    @property
    def node_to_leaf_module(self):
        return {n.physical_Id: n.moduleIndex() for n in self.infomap.iterLeafNodes()}

    @staticmethod
    def consolidate_values(adict):
        value_to_unique = {v:k for k,v in dict(enumerate(set(list(adict.values())))).items()}
        return {k: value_to_unique[v] for k,v in adict.items()}

    def group_modules(self, depth):
        module_to_cut_path = {m: self._cut_path(p, depth) for m, p in self.module_to_paths.items()}
        module_to_cut_module = self.consolidate_values(module_to_cut_path)
        return module_to_cut_module

    def _cut_path(self, path, depth):

        if len(path) <= depth:
            return path

        elif len(path)-1 == depth:
            return path[:-1]

        else:
            return self._cut_path(path[:-1], depth)

    def group(self, modules, level=0):

        if level is None:
            level = 0

        depth = self.max_depth - level - 1

        module_map = np.vectorize(self.group_modules(depth).get)

        return module_map(modules)


    # alternate more efficient method:
    # multilevel = imap.infomap.getMultilevelModules()
    # unique_paths = set([p[:depth] for p in multilevel.values()])
    # path_to_community = {path[:depth]: i for i, path  in dict(enumerate(unique_paths)).items()}
    # node_to_community = {node: path_to_community[path[:depth]] for node, path in multilevel.asdict().items()}




import numpy as np
from ..visualization import *


class GrowthTrends:
    """ Class for extracting and plotting clone trends. """

    def __init__(self, data):
        """
        Instantiate object for plotting clone trend data.

        Args:

            data (pd.DataFrame)

        """
        self.data = data.replace([np.inf, -np.inf], np.nan)
        self.x = 2**self.means.recombination_start
        self.ylabels = {
            'num_clones': 'Number of clones',
            'mean_clone_size': 'Mean clone size',
            'percent_heterozygous': 'Recombination extent',
            'transclone_edges': 'Trans-clone edges per cell',
            'clone_size_variation': 'Clone size variation'}

    @property
    def num_replicates(self):
        """ Number of growth replicates. """
        return self.data.replicate_id.max()+1

    @property
    def means(self):
        """ Mean values aggregated over replicates. """
        gb = self.data.groupby('column_id')
        return gb.mean()

    @property
    def stds(self):
        """ Std dev of values aggregated over replicates. """
        gb = self.data.groupby('column_id')
        return gb.std()

    @default_figure
    def plot_trend(self, yvar, ax=None, **kwargs):
        """ Plot <yvar> against recombination start generation. """

        # get y data
        y = self.means[yvar]
        ystd = self.stds[yvar]
        y_SEM = ystd / np.sqrt(self.num_replicates)

        # plot trend
        ax.errorbar(self.x, y, yerr=y_SEM, **kwargs)

        # format axis
        ax.set_xlabel('Recombination start (no. cells)')
        ylabel = ''
        if yvar in self.ylabels.keys():
            ylabel = self.ylabels[yvar]
        ax.set_ylabel(ylabel)
        ax.set_xscale('log', basex=2)
        ax.set_xticks([1,2,4,8,16,32,64,128,256,512])
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap

from skimage.feature import peak_local_max
from skimage.segmentation import watershed
from scipy.ndimage import distance_transform_edt, gaussian_filter
from scipy.ndimage import generate_binary_structure, iterate_structure


class Segmentation:
    """
    Object for finding nuclear contours within an image.

    Seed detection is performed by finding local maxima in a euclidean distance transform of the image foreground mask. Segmentation is achieved via the watershed method.

    Attributes:

        seeds (np.ndarray[float]) - seeds for segmentation, 2 x N

        labels (2D np.ndarray[int]) - segment label mask, number denotes segment ID

        segment_ids (1D np.ndarray[int]) - unique segment IDs, length N

        cmap (matplotlib.colors.ColorMap) - segment ID colormap, length N+1

    """

    def __init__(self, image, seed_kws={}, seg_kws={}, exclude_edges=True):
        """
        Instantiate and run segmentation.

        Args:

            image (MonochromeImage) - image to be segmented

            seed_kws (dict) - keyword arguments for seed detection

            seg_kws (dict) - keyword arguments for segmentation

            exclude_edges (bool) - if True, exclude segments on edge of image

        """
        image.set_otsu_mask()
        self.labels = None
        self.seeds = self.get_seeds_from_distance(image.mask, **seed_kws)
        self.update_cmap()
        self.watershed(image.mask, **seg_kws)

        if exclude_edges:
            self.exclude_edge_segments()

        self.segment_ids = np.array(list(self.seeds.keys()))

    @property
    def num_objects(self):
        """ Number of unique objects in image. """
        return np.unique(self.labels).size - 1

    @staticmethod
    def array_to_dict(arr):
        """ Convert array to dictionary. """
        if arr is None:
            return {}
        else:
            return {i+1: row for i, row in enumerate(arr)}

    @staticmethod
    def find_maxima(im,
                    min_distance=1,
                    num_peaks=np.inf):
        """
        Find local maxima of euclidean distance transform.

        Args:

            im (np.ndarray[bool]) - 2D boolean foreground mask

            min_distance (int) - minimum distance separating maxima, px

            num_peaks (int) - maximum number of peaks

        Returns:

            seeds (np.ndarray[float]) - local maxima, shape (N, 2)

        """
        seeds = peak_local_max(im, min_distance=min_distance, num_peaks=num_peaks, exclude_border=False)
        return seeds

    @classmethod
    def get_seeds_from_distance(cls, mask,
                                sigma=2,
                                min_distance=1,
                                num_peaks=np.inf):
        """
        Seed detection via euclidean distance transform of binary map.

        Args:

            mask (nd.ndarray[bool]) - foreground mask

            sigma (float) - smoothing applied to euclidean distance mask

            min_distance (int) - minimum pixel distance between local maxima

            num_peaks (int) - maximum number of local maxima

        Returns:

            seeds (dict) - {segment_id: (xpos, ypos)} pairs

        """

        # get values
        values = distance_transform_edt(mask).astype(float)

        # apply gaussian filter
        if sigma is not None:
            values = gaussian_filter(values, sigma)

        seeds = cls.find_maxima(values, min_distance=min_distance, num_peaks=num_peaks)
        return cls.array_to_dict(seeds)

    @classmethod
    def get_segment_mask(cls, im, seeds):
        """
        Get mask for markers.

        Args:

            im (np.ndarray[float]) - image to be segmented

            seeds (dict) - {segment_id: [x, y]} pairs

        """

        # create marker mask
        seed_mask = np.zeros_like(im, dtype=int)
        shape = np.array(seed_mask.shape).reshape(-1, 1)
        for seed_id, zyx in seeds.items():
            indices = zyx.reshape(-1, 1)
            accepted = np.alltrue((indices >= 0) & (indices < shape), axis=0)
            indices = indices[:, accepted]
            seed_mask[indices[0], indices[1]] = seed_id
        return seed_mask

    def watershed(self, mask, sigma=0.5, watershed_line=True):
        """
        Run watershed segmentation to generate segment label mask.

        Args:

            mask (np.ndarray[bool]) - binary foreground mask

            sigma (float) - parameter for smoothing distance mask

            watershed_line (bool) - if True, include 1px line between contours

        """

        # define distances
        distances = distance_transform_edt(mask)
        distances = gaussian_filter(distances, sigma=sigma)

        # run segmentation
        connectivity = iterate_structure(generate_binary_structure(2, 1), 1)
        markers = self.get_segment_mask(distances, self.seeds)
        self.labels = watershed(-distances, markers=markers, mask=mask, connectivity=connectivity, watershed_line=watershed_line)

    def update_cmap(self):
        """ Use current seeds to build colormap. """
        bg_color = np.array([[.8,.8,.8]])
        segment_colors = np.random.rand(len(self.seeds), 3)
        self.cmap = ListedColormap(np.vstack((bg_color, segment_colors)))

    @staticmethod
    def get_borders(im):
        """ Returns boolean array with borders masked as True. """
        mask = np.zeros_like(im, dtype=bool)
        mask[0, :] = True
        mask[-1, :] = True
        mask[:, 0] = True
        mask[:, -1] = True
        return mask

    def exclude_edge_segments(self):
        """ Removes segments overlaying the edge_mask. """

        # mark segments on edge of image for exclusion
        edge_segments = self.labels[self.get_borders(self.labels)]
        excluded_segments = np.unique(edge_segments)
        exclusion_mask = np.isin(self.labels, excluded_segments)

        # set edge segments to zero and remove seeds
        self.labels[exclusion_mask] = 0
        list(map(self.seeds.__delitem__, filter(self.seeds.__contains__, excluded_segments)))

    def exclude_small_segments(self, min_area=10):
        """
        Exclude small segments.

        Args:

            min_area (float) - minimum contour area

        """

        # identify small segments
        bins = np.arange(1, self.labels.max()+2)
        voxels = self.labels[self.labels!=0]
        counts, _ = np.histogram(voxels, bins=bins)

        # mark excluded segment IDs
        excluded = np.isin(np.arange(1, counts.size+1), self.segment_ids)
        excluded = np.logical_and(excluded, counts < min_area)
        excluded = bins[:-1][excluded]

        # remove small segments
        self.labels[np.isin(self.labels, excluded)] = 0
        _ = [self.seeds.pop(seed) for seed in excluded]
        self.segment_ids = np.array(list(self.seeds.keys()))

    def show(self, figsize=(15, 15)):
        """
        Visualize segment label mask.

        Args:

            figsize (tuple) - figure size

        """
        fig, ax = plt.subplots(figsize=figsize)
        ax.imshow(self.labels, cmap=self.cmap)
